/*
 *	Linux Process Unfreezer
 *
 *	Copyright (c) 1998 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

/*
 *  WARNING: Very dirty. Very experimental. Very PeeCee-centric.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <asm/page.h>
#include <asm/unistd.h>

#define LOADER_BASE 0x3ff00000
#define LOADER_TOP 0x40000000

#define STACK_SIZE PAGE_SIZE
#define STACK_TOP 0x3ff00000
#define STACK_POS (STACK_TOP - STACK_SIZE)
#define STACK_BOTTOM 0x3fe00000

#define DEF_STACK_BASE 0xa0000000
#define DEF_STACK_TOP 0xc0000000

static void __attribute__((noreturn))
die(char *msg, ...)
{
  va_list args;
  char buf[2048];

  va_start(args, msg);
  write(2, "warm: ", 6);
  strcpy(buf + vsprintf(buf, msg, args), "\n");
  write(2, buf, strlen(buf));
  exit(1);
}

static void
w_printf(char *msg, ...)
{
  va_list args;
  char buf[2048];

  va_start(args, msg);
  vsprintf(buf, msg, args);
  write(1, buf, strlen(buf));
  va_end(args);
}

static void
w_puts(char *msg)
{
  w_printf("%s\n", msg);
}

static void
chld_handler(int sig)
{
  w_puts("SIGCHLD");
}

static void
rd(int fd, void *buf, int len)
{
  if (read(fd, buf, len) != len)
    die("read: %m");
}

static int global_fd;

#define RESTORE

struct my_regs {
  unsigned long eax;
  unsigned long ebx;
  unsigned long ecx;
  unsigned long edx;
  unsigned long esi;
  unsigned long edi;
  unsigned long ebp;
  unsigned long eflags;
  unsigned long eip;
#ifdef RESTORE
  unsigned long restorer[64];
#endif
};

static void
kick_it(unsigned long addr)
{
  asm volatile ("movl %0,%%esp\n\t"
#ifdef RESTORE
#define ARG0(x) #x
#define ARG1(x) ARG0($##x)
#define ARG(z) ARG1(z)
    "            leal 36(%%esp),%%edi
                 movl $2f-1f,%%ecx\n\t\
                 movl $1f,%%esi\n\t\
                 cld\n\t\
                 rep\n\t\
                 movsb\n\t\
                 leal 36(%%esp),%%eax
                 jmp  *%%eax\n\t\
        1:	 movl " ARG(__NR_munmap) ",%%eax\n\t\
                 movl " ARG(STACK_BOTTOM) ",%%ebx\n\t\
                 movl " ARG(LOADER_TOP-STACK_BOTTOM) ",%%ecx\n\t\
                 int  $0x80\n\t\
    "
#endif
    "            popl %%eax\n\t\
                 popl %%ebx\n\t\
                 popl %%ecx\n\t\
                 popl %%edx\n\t\
                 popl %%esi\n\t\
                 popl %%edi\n\t\
                 popl %%ebp\n\t\
                 popfl\n\t"
#ifdef RESTORE
    "            ret $256\n\t\
        2:"
#else
    "            ret\n\t"
#endif
		: /* no output */ :
		"g" (addr));
}

static void
ng_main(void)
{
  unsigned long l;
  int fd = global_fd;
  struct my_regs mr;
  unsigned long esp;

  w_puts("New stack...");

  /* Unmap old stack */
  if (munmap((void *) DEF_STACK_BASE, DEF_STACK_TOP - DEF_STACK_BASE) < 0)
    die("stack munmap: %m");

  w_puts("Loading...");

  for(;;)
    {
      char buf[2048];
      int len;
      int i = read(fd, &len, sizeof(len));
      if (!i)
	break;
      if (i != sizeof(int))
	die("read: %m");
      rd(fd, buf, len);
      w_puts(buf);
      switch (buf[0])
	{
	case 'M':
	  {
	    unsigned long pos, at, len;
	    char *q;
	    int prot = 0;
	    if (buf[1] != ' ' ||
		buf[2] != 'r' ||
		(buf[3] != 'w' && buf[3] != '-') ||
		(buf[4] != 'x' && buf[4] != '-') ||
		buf[5] != ' ' ||
		sscanf(buf+6, "%lx%lx", &at, &len) != 2 ||
		!(q = strchr(buf+6, ' ')))
	      die("Format error 1");
	    if (buf[2] == 'r')
	      prot |= PROT_READ;
	    if (buf[3] == 'w')
	      prot |= PROT_WRITE;
	    if (buf[4] == 'x')
	      prot |= PROT_EXEC;
	    if (q = strchr(q+1, ' '))
	      {
		int d;
		q++;
		if (sscanf(q, "%lx", &pos) != 1 ||
		    !(q = strchr(q, ' ')))
		  die("Format error 2");
		q++;
		if ((d = open(q, O_RDONLY)) < 0)
		  die("%s: %m");
		if (mmap((void *) at, len, prot, MAP_FIXED | MAP_PRIVATE, d, pos) != (void *) at)
		  die("mmap(%s): %m", q);
		close(d);
	      }
	    else
	      {
		if (mmap((void *) at, len, prot, MAP_FIXED | MAP_PRIVATE | MAP_ANON |
			 ((at >= DEF_STACK_BASE) ? MAP_GROWSDOWN : 0), 0, 0) != (void *) at)
		  die("mmap(ANON): %m");
	      }
	    break;
	  }
	case 'W':
	  {
	    unsigned long base;
	    int size;
	    if (sscanf(buf+2, "%lx%x", &base, &size) != 2)
	      die("Format error 3");
	    rd(fd, (void *) base, size);
	    break;
	  }
	case 'R':
	  if (sscanf(buf+2, "%lx%lx%lx%lx%lx%lx%lx%lx%lx%lx",
		     &mr.eax, &mr.ebx, &mr.ecx, &mr.edx, &mr.esi, &mr.edi,
		     &mr.ebp, &esp, &mr.eip, &mr.eflags) != 10)
	    die("Format error 4");
	  break;
	default:
	  die("Unknown operation `%s'", buf);
	}
    }

  w_puts("Image loaded. Starting...");

  memcpy((void *)(esp - sizeof(mr)), &mr, sizeof(mr));
  kick_it(esp - sizeof(mr));
  die("Something is rotten.");
}

int
main(int argc, char **argv)
{
  if (argc != 2)
    {
      fprintf(stderr, "Usage: warm <file>\n");
      return 1;
    }
  if ((global_fd = open(argv[1], O_RDONLY)) < 0)
    die("%s: %m", argv[1]);

  w_printf("Shuffling stacks... (bot=%lx,stk=%lx,code=%lx)\n", (long) STACK_BOTTOM, (long) STACK_TOP, (long) LOADER_BASE);

  if (mmap((void *) STACK_POS, STACK_SIZE, PROT_READ | PROT_WRITE | PROT_EXEC,
	   MAP_PRIVATE | MAP_ANON | MAP_FIXED | MAP_GROWSDOWN, 0, 0) != (void *) STACK_POS)
    die("Cannot map new stack!");

  asm volatile("movl %0,%%esp ; call *%1" :
	       /* no output */ :
	       "g" (STACK_TOP), "r" (ng_main));
  return 23;
}
